// Created by bryankeller on 11/17/18.
// Copyright © 2018 Airbnb, Inc.

// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
// http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.

import XCTest

@testable import MagazineLayout

final class ModelStateLayoutTests: XCTestCase {

  // MARK: Internal

  override func setUp() {
    var metrics = MagazineLayoutSectionMetrics.defaultSectionMetrics(forCollectionViewWidth: 320)
    metrics.itemInsets = UIEdgeInsets(top: 10, left: 10, bottom: 10, right: 10)
    metrics.sectionInsets = UIEdgeInsets(top: 10, left: 10, bottom: 10, right: 10)
    metrics.horizontalSpacing = 20
    metrics.verticalSpacing = 30

    let sections = [
      (headerModel0, sizeModesAndHeights0, footerModel0),
      (headerModel1, sizeModesAndHeights1, footerModel1)
      ].map { headerModel, sizeModesAndHeights, footerModel in
        SectionModel(
          itemModels: sizeModesAndHeights.map { sizeMode, height in
            switch sizeMode.heightMode {
            case .static:
              return ItemModel(sizeMode: sizeMode, height: height)
            case .dynamic, .dynamicAndStretchToTallestItemInRow:
              return ItemModel(sizeMode: sizeMode, height: 150)
            }
          },
          headerModel: headerModel,
          footerModel: footerModel,
          backgroundModel: BackgroundModel(),
          metrics: metrics)
      }

    modelState = ModelState()
    modelState.setSections(sections)
  }

  override func tearDown() {
    modelState = nil
  }

  func testInitialLayout() {
    let expectedItemFrames0: [CGRect] = [
      CGRect(x: 20.0, y: 70.0, width: 280.0, height: 20.0),
      CGRect(x: 0.0, y: 120.0, width: 320.0, height: 150.0),
      CGRect(x: 20.0, y: 300.0, width: 130.0, height: 10.0),
      CGRect(x: 170.0, y: 300.0, width: 130.0, height: 30.0),
      CGRect(x: 20.0, y: 360.0, width: 130.0, height: 150.0),
    ]
    let expectedItemFrames1: [CGRect] = [
      CGRect(x: 20.0, y: 360.0, width: 130.0, height: 150.0),
      CGRect(x: 20.0, y: 540.0, width: 80.0, height: 150.0),
      CGRect(x: 120.0, y: 540.0, width: 80.0, height: 150.0),
      CGRect(x: 20.0, y: 720.0, width: 55.0, height: 15.0),
      CGRect(x: 95.0, y: 720.0, width: 55.0, height: 150.0),
      CGRect(x: 170.0, y: 720.0, width: 55.0, height: 150.0),
      CGRect(x: 245.0, y: 720.0, width: 55.0, height: 150.0),
      CGRect(x: 20.0, y: 900.0, width: 40.0, height: 150.0),
      CGRect(x: 20.0, y: 1210.0, width: 130.0, height: 10.0),
      CGRect(x: 170.0, y: 1210.0, width: 130.0, height: 30.0),
      CGRect(x: 20.0, y: 1270.0, width: 130.0, height: 25.0),
      CGRect(x: 20.0, y: 1325.0, width: 80.0, height: 15.0),
      CGRect(x: 20.0, y: 1370.0, width: 280.0, height: 20.0),
      CGRect(x: 20.0, y: 1420.0, width: 80.0, height: 10.0),
      CGRect(x: 20.0, y: 1460.0, width: 40.0, height: 15.0),
      CGRect(x: 80.0, y: 1460.0, width: 40.0, height: 15.0),
      CGRect(x: 140.0, y: 1460.0, width: 40.0, height: 25.0),
      CGRect(x: 200.0, y: 1460.0, width: 40.0, height: 35.0),
      CGRect(x: 260.0, y: 1460.0, width: 40.0, height: 30.0),
      CGRect(x: 0.0, y: 1525.0, width: 320.0, height: 15.0),
    ]
    let expectedHeaderFrames0: [CGRect] = [
      CGRect(x: 10.0, y: 10.0, width: 300.0, height: 50.0),
    ]
    let expectedHeaderFrames1: [CGRect] = [
      CGRect(x: 10.0, y: 1130.0, width: 300.0, height: 70.0),
    ]
    let expectedFooterFrames0 = [CGRect]()
    let expectedFooterFrames1: [CGRect] = [
      CGRect(x: 10.0, y: 1060.0, width: 300.0, height: 50.0),
      CGRect(x: 10.0, y: 1550.0, width: 300.0, height: 70.0),
    ]
    let expectedBackgroundFrames0: [CGRect] = [
      CGRect(x: 10.0, y: 10.0, width: 300.0, height: 1110.0),
    ]
    let expectedBackgroundFrames1: [CGRect] = [
      CGRect(x: 10.0, y: 10.0, width: 300.0, height: 1110.0),
      CGRect(x: 10.0, y: 1130.0, width: 300.0, height: 500.0),
    ]

    checkExpectedFrames(
      expectedItemFrames0: expectedItemFrames0,
      expectedItemFrames1: expectedItemFrames1,
      expectedHeaderFrames0: expectedHeaderFrames0,
      expectedHeaderFrames1: expectedHeaderFrames1,
      expectedFooterFrames0: expectedFooterFrames0,
      expectedFooterFrames1: expectedFooterFrames1,
      expectedBackgroundFrames0: expectedBackgroundFrames0,
      expectedBackgroundFrames1: expectedBackgroundFrames1)
  }

  func testUpdatingPreferredHeights() {
    modelState.updateHeaderHeight(toPreferredHeight: 50, forSectionAtIndex: 1)

    for (itemIndex, sizeModeAndHeight) in sizeModesAndHeights0.enumerated() {
      switch sizeModeAndHeight.sizeMode.heightMode {
      case .dynamic, .dynamicAndStretchToTallestItemInRow:
        let indexPath = IndexPath(item: itemIndex, section: 0)
        modelState.updateItemHeight(
          toPreferredHeight: sizeModeAndHeight.height,
          forItemAt: indexPath)

        let preferredHeight = modelState.itemModelPreferredHeightDuringPreferredAttributesCheck(
          at: indexPath)
        XCTAssert(preferredHeight == sizeModeAndHeight.height, "Item preferred height is incorrect")
      case .static:
        continue
      }
    }

    modelState.updateFooterHeight(toPreferredHeight: 50, forSectionAtIndex: 1)

    let expectedItemFrames0: [CGRect] = [
      CGRect(x: 20.0, y: 70.0, width: 280.0, height: 20.0),
      CGRect(x: 0.0, y: 120.0, width: 320.0, height: 15.0),
      CGRect(x: 20.0, y: 165.0, width: 130.0, height: 10.0),
      CGRect(x: 170.0, y: 165.0, width: 130.0, height: 30.0),
      CGRect(x: 20.0, y: 225.0, width: 130.0, height: 25.0),
      CGRect(x: 20.0, y: 280.0, width: 80.0, height: 15.0),
      CGRect(x: 120.0, y: 280.0, width: 80.0, height: 15.0),
      CGRect(x: 20.0, y: 325.0, width: 55.0, height: 15.0),
      CGRect(x: 95.0, y: 325.0, width: 55.0, height: 15.0),
      CGRect(x: 170.0, y: 325.0, width: 55.0, height: 25.0),
      CGRect(x: 245.0, y: 325.0, width: 55.0, height: 35.0),
      CGRect(x: 20.0, y: 390.0, width: 40.0, height: 30.0),
    ]
    let expectedItemFrames1: [CGRect] = [
      CGRect(x: 20.0, y: 560.0, width: 130.0, height: 10.0),
      CGRect(x: 170.0, y: 560.0, width: 130.0, height: 30.0),
      CGRect(x: 20.0, y: 620.0, width: 130.0, height: 25.0),
      CGRect(x: 20.0, y: 675.0, width: 80.0, height: 15.0),
      CGRect(x: 20.0, y: 720.0, width: 280.0, height: 20.0),
      CGRect(x: 20.0, y: 770.0, width: 80.0, height: 10.0),
      CGRect(x: 20.0, y: 810.0, width: 40.0, height: 15.0),
      CGRect(x: 80.0, y: 810.0, width: 40.0, height: 15.0),
      CGRect(x: 140.0, y: 810.0, width: 40.0, height: 25.0),
      CGRect(x: 200.0, y: 810.0, width: 40.0, height: 35.0),
      CGRect(x: 260.0, y: 810.0, width: 40.0, height: 30.0),
      CGRect(x: 0.0, y: 875.0, width: 320.0, height: 15.0),
    ]
    let expectedHeaderFrames0: [CGRect] = [
      CGRect(x: 10.0, y: 10.0, width: 300.0, height: 50.0),
    ]
    let expectedHeaderFrames1: [CGRect] = [
      CGRect(x: 10.0, y: 500.0, width: 300.0, height: 50.0),
    ]
    let expectedFooterFrames0: [CGRect] = [
      CGRect(x: 10.0, y: 430.0, width: 300.0, height: 50.0),
    ]
    let expectedFooterFrames1: [CGRect] = [
      CGRect(x: 10.0, y: 900.0, width: 300.0, height: 50.0),
    ]
    let expectedBackgroundFrames0: [CGRect] = [
      CGRect(x: 10.0, y: 10.0, width: 300.0, height: 480.0),
    ]
    let expectedBackgroundFrames1: [CGRect] = [
      CGRect(x: 10.0, y: 500.0, width: 300.0, height: 460.0),
    ]

    checkExpectedFrames(
      expectedItemFrames0: expectedItemFrames0,
      expectedItemFrames1: expectedItemFrames1,
      expectedHeaderFrames0: expectedHeaderFrames0,
      expectedHeaderFrames1: expectedHeaderFrames1,
      expectedFooterFrames0: expectedFooterFrames0,
      expectedFooterFrames1: expectedFooterFrames1,
      expectedBackgroundFrames0: expectedBackgroundFrames0,
      expectedBackgroundFrames1: expectedBackgroundFrames1)
  }

  func testUpdatingSectionMetrics() {
    var metrics = MagazineLayoutSectionMetrics.defaultSectionMetrics(forCollectionViewWidth: 100)
    metrics.itemInsets = UIEdgeInsets(top: 2, left: 2, bottom: 2, right: 2)
    metrics.sectionInsets = UIEdgeInsets(top: 2, left: 2, bottom: 2, right: 2)
    metrics.horizontalSpacing = 5
    metrics.verticalSpacing = 100
    modelState.updateMetrics(to: metrics, forSectionAtIndex: 0)

    let expectedItemFrames0: [CGRect] = [
      CGRect(x: 4.0, y: 54.0, width: 92.0, height: 20.0),
      CGRect(x: 0.0, y: 174.0, width: 100.0, height: 150.0),
      CGRect(x: 4.0, y: 424.0, width: 44.0, height: 10.0),
      CGRect(x: 53.0, y: 424.0, width: 44.0, height: 30.0),
    ]
    let expectedItemFrames1: [CGRect] = [
      CGRect(x: 4.0, y: 554.0, width: 44.0, height: 150.0),
      CGRect(x: 4.0, y: 804.0, width: 27.0, height: 150.0),
      CGRect(x: 36.0, y: 804.0, width: 27.0, height: 150.0),
      CGRect(x: 4.0, y: 1054.0, width: 19.0, height: 15.0),
      CGRect(x: 28.0, y: 1054.0, width: 19.0, height: 150.0),
      CGRect(x: 52.0, y: 1054.0, width: 19.0, height: 150.0),
      CGRect(x: 76.0, y: 1054.0, width: 19.0, height: 150.0),
      CGRect(x: 4.0, y: 1304.0, width: 14.0, height: 150.0),
      CGRect(x: 20.0, y: 1598.0, width: 130.0, height: 10.0),
      CGRect(x: 170.0, y: 1598.0, width: 130.0, height: 30.0),
      CGRect(x: 20.0, y: 1658.0, width: 130.0, height: 25.0),
      CGRect(x: 20.0, y: 1713.0, width: 80.0, height: 15.0),
      CGRect(x: 20.0, y: 1758.0, width: 280.0, height: 20.0),
      CGRect(x: 20.0, y: 1808.0, width: 80.0, height: 10.0),
      CGRect(x: 20.0, y: 1848.0, width: 40.0, height: 15.0),
      CGRect(x: 80.0, y: 1848.0, width: 40.0, height: 15.0),
      CGRect(x: 140.0, y: 1848.0, width: 40.0, height: 25.0),
      CGRect(x: 200.0, y: 1848.0, width: 40.0, height: 35.0),
      CGRect(x: 260.0, y: 1848.0, width: 40.0, height: 30.0),
      CGRect(x: 0.0, y: 1913.0, width: 320.0, height: 15.0),
    ]
    let expectedHeaderFrames0: [CGRect] = [
      CGRect(x: 2.0, y: 2.0, width: 96.0, height: 50.0),
    ]
    let expectedHeaderFrames1: [CGRect] = [
      CGRect(x: 10.0, y: 1518.0, width: 300.0, height: 70.0),
    ]
    let expectedFooterFrames0 = [CGRect]()
    let expectedFooterFrames1: [CGRect] = [
      CGRect(x: 2.0, y: 1456.0, width: 96.0, height: 50.0),
      CGRect(x: 10.0, y: 1938.0, width: 300.0, height: 70.0),
    ]
    let expectedBackgroundFrames0: [CGRect] = [
      CGRect(x: 2.0, y: 2.0, width: 96.0, height: 1506.0),
    ]
    let expectedBackgroundFrames1: [CGRect] = [
      CGRect(x: 2.0, y: 2.0, width: 96.0, height: 1506.0),
      CGRect(x: 10.0, y: 1518.0, width: 300.0, height: 500.0),
    ]

    checkExpectedFrames(
      expectedItemFrames0: expectedItemFrames0,
      expectedItemFrames1: expectedItemFrames1,
      expectedHeaderFrames0: expectedHeaderFrames0,
      expectedHeaderFrames1: expectedHeaderFrames1,
      expectedFooterFrames0: expectedFooterFrames0,
      expectedFooterFrames1: expectedFooterFrames1,
      expectedBackgroundFrames0: expectedBackgroundFrames0,
      expectedBackgroundFrames1: expectedBackgroundFrames1)
  }

  func testUpdatingItemSizeMode() {
    modelState.updateItemSizeMode(
      to: MagazineLayoutItemSizeMode(
        widthMode: .thirdWidth,
        heightMode: .static(height: 50)),
      forItemAt: IndexPath(item: 5, section: 0))
    modelState.updateItemSizeMode(
      to: MagazineLayoutItemSizeMode(
        widthMode: .fifthWidth,
        heightMode: .dynamicAndStretchToTallestItemInRow),
      forItemAt: IndexPath(item: 8, section: 1))

    let expectedItemFrames0: [CGRect] = [
      CGRect(x: 20.0, y: 70.0, width: 280.0, height: 20.0),
      CGRect(x: 0.0, y: 120.0, width: 320.0, height: 150.0),
      CGRect(x: 20.0, y: 300.0, width: 130.0, height: 10.0),
      CGRect(x: 170.0, y: 300.0, width: 130.0, height: 30.0),
      CGRect(x: 20.0, y: 360.0, width: 130.0, height: 150.0),
    ]
    let expectedItemFrames1: [CGRect] = [
      CGRect(x: 20.0, y: 360.0, width: 130.0, height: 150.0),
      CGRect(x: 20.0, y: 540.0, width: 80.0, height: 50.0),
      CGRect(x: 120.0, y: 540.0, width: 80.0, height: 150.0),
      CGRect(x: 20.0, y: 720.0, width: 55.0, height: 15.0),
      CGRect(x: 95.0, y: 720.0, width: 55.0, height: 150.0),
      CGRect(x: 170.0, y: 720.0, width: 55.0, height: 150.0),
      CGRect(x: 245.0, y: 720.0, width: 55.0, height: 150.0),
      CGRect(x: 20.0, y: 900.0, width: 40.0, height: 150.0),
      CGRect(x: 20.0, y: 1210.0, width: 130.0, height: 10.0),
      CGRect(x: 170.0, y: 1210.0, width: 130.0, height: 30.0),
      CGRect(x: 20.0, y: 1270.0, width: 130.0, height: 25.0),
      CGRect(x: 20.0, y: 1325.0, width: 80.0, height: 15.0),
      CGRect(x: 20.0, y: 1370.0, width: 280.0, height: 20.0),
      CGRect(x: 20.0, y: 1420.0, width: 80.0, height: 10.0),
      CGRect(x: 20.0, y: 1460.0, width: 40.0, height: 15.0),
      CGRect(x: 80.0, y: 1460.0, width: 40.0, height: 15.0),
      CGRect(x: 140.0, y: 1460.0, width: 40.0, height: 35.0),
      CGRect(x: 200.0, y: 1460.0, width: 40.0, height: 35.0),
      CGRect(x: 260.0, y: 1460.0, width: 40.0, height: 30.0),
      CGRect(x: 0.0, y: 1525.0, width: 320.0, height: 15.0),
    ]
    let expectedHeaderFrames0: [CGRect] = [
      CGRect(x: 10.0, y: 10.0, width: 300.0, height: 50.0),
    ]
    let expectedHeaderFrames1: [CGRect] = [
      CGRect(x: 10.0, y: 1130.0, width: 300.0, height: 70.0),
    ]
    let expectedFooterFrames0 = [CGRect]()
    let expectedFooterFrames1: [CGRect] = [
      CGRect(x: 10.0, y: 1060.0, width: 300.0, height: 50.0),
      CGRect(x: 10.0, y: 1550.0, width: 300.0, height: 70.0),
    ]
    let expectedBackgroundFrames0: [CGRect] = [
      CGRect(x: 10.0, y: 10.0, width: 300.0, height: 1110.0),
    ]
    let expectedBackgroundFrames1: [CGRect] = [
      CGRect(x: 10.0, y: 10.0, width: 300.0, height: 1110.0),
      CGRect(x: 10.0, y: 1130.0, width: 300.0, height: 500.0),
    ]

    checkExpectedFrames(
      expectedItemFrames0: expectedItemFrames0,
      expectedItemFrames1: expectedItemFrames1,
      expectedHeaderFrames0: expectedHeaderFrames0,
      expectedHeaderFrames1: expectedHeaderFrames1,
      expectedFooterFrames0: expectedFooterFrames0,
      expectedFooterFrames1: expectedFooterFrames1,
      expectedBackgroundFrames0: expectedBackgroundFrames0,
      expectedBackgroundFrames1: expectedBackgroundFrames1)
  }

  func testReplacingHeader() {
    modelState.removeHeader(forSectionAtIndex: 0)

    let expectedItemFrames0: [CGRect] = [
      CGRect(x: 20.0, y: 10.0, width: 280.0, height: 20.0),
      CGRect(x: 0.0, y: 60.0, width: 320.0, height: 150.0),
      CGRect(x: 20.0, y: 240.0, width: 130.0, height: 10.0),
      CGRect(x: 170.0, y: 240.0, width: 130.0, height: 30.0),
      CGRect(x: 20.0, y: 300.0, width: 130.0, height: 150.0),
      CGRect(x: 20.0, y: 480.0, width: 80.0, height: 150.0),
      CGRect(x: 120.0, y: 480.0, width: 80.0, height: 150.0),
    ]
    let expectedItemFrames1: [CGRect] = [
      CGRect(x: 20.0, y: 480.0, width: 80.0, height: 150.0),
      CGRect(x: 120.0, y: 480.0, width: 80.0, height: 150.0),
      CGRect(x: 20.0, y: 660.0, width: 55.0, height: 15.0),
      CGRect(x: 95.0, y: 660.0, width: 55.0, height: 150.0),
      CGRect(x: 170.0, y: 660.0, width: 55.0, height: 150.0),
      CGRect(x: 245.0, y: 660.0, width: 55.0, height: 150.0),
      CGRect(x: 20.0, y: 840.0, width: 40.0, height: 150.0),
      CGRect(x: 20.0, y: 1150.0, width: 130.0, height: 10.0),
      CGRect(x: 170.0, y: 1150.0, width: 130.0, height: 30.0),
      CGRect(x: 20.0, y: 1210.0, width: 130.0, height: 25.0),
      CGRect(x: 20.0, y: 1265.0, width: 80.0, height: 15.0),
      CGRect(x: 20.0, y: 1310.0, width: 280.0, height: 20.0),
      CGRect(x: 20.0, y: 1360.0, width: 80.0, height: 10.0),
      CGRect(x: 20.0, y: 1400.0, width: 40.0, height: 15.0),
      CGRect(x: 80.0, y: 1400.0, width: 40.0, height: 15.0),
      CGRect(x: 140.0, y: 1400.0, width: 40.0, height: 25.0),
      CGRect(x: 200.0, y: 1400.0, width: 40.0, height: 35.0),
      CGRect(x: 260.0, y: 1400.0, width: 40.0, height: 30.0),
      CGRect(x: 0.0, y: 1465.0, width: 320.0, height: 15.0),
    ]
    let expectedHeaderFrames0: [CGRect] = []
    let expectedHeaderFrames1: [CGRect] = [
      CGRect(x: 10.0, y: 1070.0, width: 300.0, height: 70.0),
    ]
    let expectedFooterFrames0 = [CGRect]()
    let expectedFooterFrames1: [CGRect] = [
      CGRect(x: 10.0, y: 1000.0, width: 300.0, height: 50.0),
      CGRect(x: 10.0, y: 1490.0, width: 300.0, height: 70.0),
    ]
    let expectedBackgroundFrames0: [CGRect] = [
      CGRect(x: 10.0, y: 10.0, width: 300.0, height: 1050.0),
    ]
    let expectedBackgroundFrames1: [CGRect] = [
      CGRect(x: 10.0, y: 10.0, width: 300.0, height: 1050.0),
      CGRect(x: 10.0, y: 1070.0, width: 300.0, height: 500.0),
    ]

    checkExpectedFrames(
      expectedItemFrames0: expectedItemFrames0,
      expectedItemFrames1: expectedItemFrames1,
      expectedHeaderFrames0: expectedHeaderFrames0,
      expectedHeaderFrames1: expectedHeaderFrames1,
      expectedFooterFrames0: expectedFooterFrames0,
      expectedFooterFrames1: expectedFooterFrames1,
      expectedBackgroundFrames0: expectedBackgroundFrames0,
      expectedBackgroundFrames1: expectedBackgroundFrames1)

    modelState.setHeader(
      HeaderModel(heightMode: .static(height: 60), height: 60),
      forSectionAtIndex: 0)
    modelState.setHeader(
      HeaderModel(heightMode: .dynamic, height: 100),
      forSectionAtIndex: 1)

    let expectedItemFrames2: [CGRect] = [
      CGRect(x: 20.0, y: 80.0, width: 280.0, height: 20.0),
      CGRect(x: 0.0, y: 130.0, width: 320.0, height: 150.0),
      CGRect(x: 20.0, y: 310.0, width: 130.0, height: 10.0),
      CGRect(x: 170.0, y: 310.0, width: 130.0, height: 30.0),
      CGRect(x: 20.0, y: 370.0, width: 130.0, height: 150.0),
    ]
    let expectedItemFrames3: [CGRect] = [
      CGRect(x: 20.0, y: 370.0, width: 130.0, height: 150.0),
      CGRect(x: 20.0, y: 550.0, width: 80.0, height: 150.0),
      CGRect(x: 120.0, y: 550.0, width: 80.0, height: 150.0),
      CGRect(x: 20.0, y: 730.0, width: 55.0, height: 15.0),
      CGRect(x: 95.0, y: 730.0, width: 55.0, height: 150.0),
      CGRect(x: 170.0, y: 730.0, width: 55.0, height: 150.0),
      CGRect(x: 245.0, y: 730.0, width: 55.0, height: 150.0),
      CGRect(x: 20.0, y: 910.0, width: 40.0, height: 150.0),
      CGRect(x: 20.0, y: 1250.0, width: 130.0, height: 10.0),
      CGRect(x: 170.0, y: 1250.0, width: 130.0, height: 30.0),
      CGRect(x: 20.0, y: 1310.0, width: 130.0, height: 25.0),
      CGRect(x: 20.0, y: 1365.0, width: 80.0, height: 15.0),
      CGRect(x: 20.0, y: 1410.0, width: 280.0, height: 20.0),
      CGRect(x: 20.0, y: 1460.0, width: 80.0, height: 10.0),
      CGRect(x: 20.0, y: 1500.0, width: 40.0, height: 15.0),
      CGRect(x: 80.0, y: 1500.0, width: 40.0, height: 15.0),
      CGRect(x: 140.0, y: 1500.0, width: 40.0, height: 25.0),
      CGRect(x: 200.0, y: 1500.0, width: 40.0, height: 35.0),
      CGRect(x: 260.0, y: 1500.0, width: 40.0, height: 30.0),
      CGRect(x: 0.0, y: 1565.0, width: 320.0, height: 15.0),
    ]
    let expectedHeaderFrames2: [CGRect] = [
      CGRect(x: 10.0, y: 10.0, width: 300.0, height: 60.0),
    ]
    let expectedHeaderFrames3: [CGRect] = [
      CGRect(x: 10.0, y: 1140.0, width: 300.0, height: 100.0),
    ]
    let expectedFooterFrames2 = [CGRect]()
    let expectedFooterFrames3: [CGRect] = [
      CGRect(x: 10.0, y: 1070.0, width: 300.0, height: 50.0),
      CGRect(x: 10.0, y: 1590.0, width: 300.0, height: 70.0),
    ]
    let expectedBackgroundFrames2: [CGRect] = [
      CGRect(x: 10.0, y: 10.0, width: 300.0, height: 1120.0),
    ]
    let expectedBackgroundFrames3: [CGRect] = [
      CGRect(x: 10.0, y: 10.0, width: 300.0, height: 1120.0),
      CGRect(x: 10.0, y: 1140.0, width: 300.0, height: 530.0),
    ]

    checkExpectedFrames(
      expectedItemFrames0: expectedItemFrames2,
      expectedItemFrames1: expectedItemFrames3,
      expectedHeaderFrames0: expectedHeaderFrames2,
      expectedHeaderFrames1: expectedHeaderFrames3,
      expectedFooterFrames0: expectedFooterFrames2,
      expectedFooterFrames1: expectedFooterFrames3,
      expectedBackgroundFrames0: expectedBackgroundFrames2,
      expectedBackgroundFrames1: expectedBackgroundFrames3)
  }

  func testReplacingFooters() {
    modelState.removeFooter(forSectionAtIndex: 0)

    let expectedItemFrames0: [CGRect] = [
      CGRect(x: 20.0, y: 70.0, width: 280.0, height: 20.0),
      CGRect(x: 0.0, y: 120.0, width: 320.0, height: 150.0),
      CGRect(x: 20.0, y: 300.0, width: 130.0, height: 10.0),
      CGRect(x: 170.0, y: 300.0, width: 130.0, height: 30.0),
      CGRect(x: 20.0, y: 360.0, width: 130.0, height: 150.0),
    ]
    let expectedItemFrames1: [CGRect] = [
      CGRect(x: 20.0, y: 360.0, width: 130.0, height: 150.0),
      CGRect(x: 20.0, y: 540.0, width: 80.0, height: 150.0),
      CGRect(x: 120.0, y: 540.0, width: 80.0, height: 150.0),
      CGRect(x: 20.0, y: 720.0, width: 55.0, height: 15.0),
      CGRect(x: 95.0, y: 720.0, width: 55.0, height: 150.0),
      CGRect(x: 170.0, y: 720.0, width: 55.0, height: 150.0),
      CGRect(x: 245.0, y: 720.0, width: 55.0, height: 150.0),
      CGRect(x: 20.0, y: 900.0, width: 40.0, height: 150.0),
      CGRect(x: 20.0, y: 1210.0, width: 130.0, height: 10.0),
      CGRect(x: 170.0, y: 1210.0, width: 130.0, height: 30.0),
      CGRect(x: 20.0, y: 1270.0, width: 130.0, height: 25.0),
      CGRect(x: 20.0, y: 1325.0, width: 80.0, height: 15.0),
      CGRect(x: 20.0, y: 1370.0, width: 280.0, height: 20.0),
      CGRect(x: 20.0, y: 1420.0, width: 80.0, height: 10.0),
      CGRect(x: 20.0, y: 1460.0, width: 40.0, height: 15.0),
      CGRect(x: 80.0, y: 1460.0, width: 40.0, height: 15.0),
      CGRect(x: 140.0, y: 1460.0, width: 40.0, height: 25.0),
      CGRect(x: 200.0, y: 1460.0, width: 40.0, height: 35.0),
      CGRect(x: 260.0, y: 1460.0, width: 40.0, height: 30.0),
      CGRect(x: 0.0, y: 1525.0, width: 320.0, height: 15.0),
    ]
    let expectedHeaderFrames0: [CGRect] = [
      CGRect(x: 10.0, y: 10.0, width: 300.0, height: 50.0),
    ]
    let expectedHeaderFrames1: [CGRect] = [
      CGRect(x: 10.0, y: 1130.0, width: 300.0, height: 70.0),
    ]
    let expectedFooterFrames0 = [CGRect]()
    let expectedFooterFrames1: [CGRect] = [
      CGRect(x: 10.0, y: 1550.0, width: 300.0, height: 70.0),
    ]
    let expectedBackgroundFrames0: [CGRect] = [
      CGRect(x: 10.0, y: 10.0, width: 300.0, height: 1110.0),
    ]
    let expectedBackgroundFrames1: [CGRect] = [
      CGRect(x: 10.0, y: 10.0, width: 300.0, height: 1110.0),
      CGRect(x: 10.0, y: 1130.0, width: 300.0, height: 500.0),
    ]

    checkExpectedFrames(
      expectedItemFrames0: expectedItemFrames0,
      expectedItemFrames1: expectedItemFrames1,
      expectedHeaderFrames0: expectedHeaderFrames0,
      expectedHeaderFrames1: expectedHeaderFrames1,
      expectedFooterFrames0: expectedFooterFrames0,
      expectedFooterFrames1: expectedFooterFrames1,
      expectedBackgroundFrames0: expectedBackgroundFrames0,
      expectedBackgroundFrames1: expectedBackgroundFrames1)

    modelState.setFooter(
      FooterModel(heightMode: .static(height: 60), height: 60),
      forSectionAtIndex: 0)
    modelState.setFooter(
      FooterModel(heightMode: .dynamic, height: 100),
      forSectionAtIndex: 1)

    let expectedItemFrames2: [CGRect] = [
      CGRect(x: 20.0, y: 70.0, width: 280.0, height: 20.0),
      CGRect(x: 0.0, y: 120.0, width: 320.0, height: 150.0),
      CGRect(x: 20.0, y: 300.0, width: 130.0, height: 10.0),
      CGRect(x: 170.0, y: 300.0, width: 130.0, height: 30.0),
      CGRect(x: 20.0, y: 360.0, width: 130.0, height: 150.0),
    ]
    let expectedItemFrames3: [CGRect] = [
      CGRect(x: 20.0, y: 360.0, width: 130.0, height: 150.0),
      CGRect(x: 20.0, y: 540.0, width: 80.0, height: 150.0),
      CGRect(x: 120.0, y: 540.0, width: 80.0, height: 150.0),
      CGRect(x: 20.0, y: 720.0, width: 55.0, height: 15.0),
      CGRect(x: 95.0, y: 720.0, width: 55.0, height: 150.0),
      CGRect(x: 170.0, y: 720.0, width: 55.0, height: 150.0),
      CGRect(x: 245.0, y: 720.0, width: 55.0, height: 150.0),
      CGRect(x: 20.0, y: 900.0, width: 40.0, height: 150.0),
      CGRect(x: 20.0, y: 1220.0, width: 130.0, height: 10.0),
      CGRect(x: 170.0, y: 1220.0, width: 130.0, height: 30.0),
      CGRect(x: 20.0, y: 1280.0, width: 130.0, height: 25.0),
      CGRect(x: 20.0, y: 1335.0, width: 80.0, height: 15.0),
      CGRect(x: 20.0, y: 1380.0, width: 280.0, height: 20.0),
      CGRect(x: 20.0, y: 1430.0, width: 80.0, height: 10.0),
      CGRect(x: 20.0, y: 1470.0, width: 40.0, height: 15.0),
      CGRect(x: 80.0, y: 1470.0, width: 40.0, height: 15.0),
      CGRect(x: 140.0, y: 1470.0, width: 40.0, height: 25.0),
      CGRect(x: 200.0, y: 1470.0, width: 40.0, height: 35.0),
      CGRect(x: 260.0, y: 1470.0, width: 40.0, height: 30.0),
      CGRect(x: 0.0, y: 1535.0, width: 320.0, height: 15.0),
    ]
    let expectedHeaderFrames2: [CGRect] = [
      CGRect(x: 10.0, y: 10.0, width: 300.0, height: 50.0),
    ]
    let expectedHeaderFrames3: [CGRect] = [
      CGRect(x: 10.0, y: 1140.0, width: 300.0, height: 70.0),
    ]
    let expectedFooterFrames2: [CGRect] = [
      CGRect(x: 0.0, y: 0.0, width: 0.0, height: 60.0),
    ]
    let expectedFooterFrames3: [CGRect] = [
      CGRect(x: 0.0, y: 1130.0, width: 0.0, height: 100.0),
    ]
    let expectedBackgroundFrames2: [CGRect] = [
      CGRect(x: 10.0, y: 10.0, width: 300.0, height: 1120.0),
    ]
    let expectedBackgroundFrames3: [CGRect] = [
      CGRect(x: 10.0, y: 10.0, width: 300.0, height: 1120.0),
      CGRect(x: 10.0, y: 1140.0, width: 300.0, height: 530.0),
    ]

    checkExpectedFrames(
      expectedItemFrames0: expectedItemFrames2,
      expectedItemFrames1: expectedItemFrames3,
      expectedHeaderFrames0: expectedHeaderFrames2,
      expectedHeaderFrames1: expectedHeaderFrames3,
      expectedFooterFrames0: expectedFooterFrames2,
      expectedFooterFrames1: expectedFooterFrames3,
      expectedBackgroundFrames0: expectedBackgroundFrames2,
      expectedBackgroundFrames1: expectedBackgroundFrames3)
  }

  func testReplacingBackground() {
    modelState.removeBackground(forSectionAtIndex: 0)

    let expectedItemFrames0: [CGRect] = [
      CGRect(x: 20.0, y: 70.0, width: 280.0, height: 20.0),
      CGRect(x: 0.0, y: 120.0, width: 320.0, height: 150.0),
      CGRect(x: 20.0, y: 300.0, width: 130.0, height: 10.0),
      CGRect(x: 170.0, y: 300.0, width: 130.0, height: 30.0),
      CGRect(x: 20.0, y: 360.0, width: 130.0, height: 150.0),
    ]
    let expectedItemFrames1: [CGRect] = [
      CGRect(x: 20.0, y: 360.0, width: 130.0, height: 150.0),
      CGRect(x: 20.0, y: 540.0, width: 80.0, height: 150.0),
      CGRect(x: 120.0, y: 540.0, width: 80.0, height: 150.0),
      CGRect(x: 20.0, y: 720.0, width: 55.0, height: 15.0),
      CGRect(x: 95.0, y: 720.0, width: 55.0, height: 150.0),
      CGRect(x: 170.0, y: 720.0, width: 55.0, height: 150.0),
      CGRect(x: 245.0, y: 720.0, width: 55.0, height: 150.0),
      CGRect(x: 20.0, y: 900.0, width: 40.0, height: 150.0),
      CGRect(x: 20.0, y: 1210.0, width: 130.0, height: 10.0),
      CGRect(x: 170.0, y: 1210.0, width: 130.0, height: 30.0),
      CGRect(x: 20.0, y: 1270.0, width: 130.0, height: 25.0),
      CGRect(x: 20.0, y: 1325.0, width: 80.0, height: 15.0),
      CGRect(x: 20.0, y: 1370.0, width: 280.0, height: 20.0),
      CGRect(x: 20.0, y: 1420.0, width: 80.0, height: 10.0),
      CGRect(x: 20.0, y: 1460.0, width: 40.0, height: 15.0),
      CGRect(x: 80.0, y: 1460.0, width: 40.0, height: 15.0),
      CGRect(x: 140.0, y: 1460.0, width: 40.0, height: 25.0),
      CGRect(x: 200.0, y: 1460.0, width: 40.0, height: 35.0),
      CGRect(x: 260.0, y: 1460.0, width: 40.0, height: 30.0),
      CGRect(x: 0.0, y: 1525.0, width: 320.0, height: 15.0),
    ]
    let expectedHeaderFrames0: [CGRect] = [
      CGRect(x: 10.0, y: 10.0, width: 300.0, height: 50.0),
    ]
    let expectedHeaderFrames1: [CGRect] = [
      CGRect(x: 10.0, y: 1130.0, width: 300.0, height: 70.0),
    ]
    let expectedFooterFrames0 = [CGRect]()
    let expectedFooterFrames1: [CGRect] = [
      CGRect(x: 10.0, y: 1060.0, width: 300.0, height: 50.0),
      CGRect(x: 10.0, y: 1550.0, width: 300.0, height: 70.0),
    ]
    let expectedBackgroundFrames0 = [CGRect]()
    let expectedBackgroundFrames1: [CGRect] = [
      CGRect(x: 10.0, y: 1130.0, width: 300.0, height: 500.0),
    ]

    checkExpectedFrames(
      expectedItemFrames0: expectedItemFrames0,
      expectedItemFrames1: expectedItemFrames1,
      expectedHeaderFrames0: expectedHeaderFrames0,
      expectedHeaderFrames1: expectedHeaderFrames1,
      expectedFooterFrames0: expectedFooterFrames0,
      expectedFooterFrames1: expectedFooterFrames1,
      expectedBackgroundFrames0: expectedBackgroundFrames0,
      expectedBackgroundFrames1: expectedBackgroundFrames1)

    modelState.setBackground(BackgroundModel(), forSectionAtIndex: 0)

    let expectedItemFrames2: [CGRect] = [
      CGRect(x: 20.0, y: 70.0, width: 280.0, height: 20.0),
      CGRect(x: 0.0, y: 120.0, width: 320.0, height: 150.0),
      CGRect(x: 20.0, y: 300.0, width: 130.0, height: 10.0),
      CGRect(x: 170.0, y: 300.0, width: 130.0, height: 30.0),
      CGRect(x: 20.0, y: 360.0, width: 130.0, height: 150.0),
    ]
    let expectedItemFrames3: [CGRect] = [
      CGRect(x: 20.0, y: 360.0, width: 130.0, height: 150.0),
      CGRect(x: 20.0, y: 540.0, width: 80.0, height: 150.0),
      CGRect(x: 120.0, y: 540.0, width: 80.0, height: 150.0),
      CGRect(x: 20.0, y: 720.0, width: 55.0, height: 15.0),
      CGRect(x: 95.0, y: 720.0, width: 55.0, height: 150.0),
      CGRect(x: 170.0, y: 720.0, width: 55.0, height: 150.0),
      CGRect(x: 245.0, y: 720.0, width: 55.0, height: 150.0),
      CGRect(x: 20.0, y: 900.0, width: 40.0, height: 150.0),
      CGRect(x: 20.0, y: 1210.0, width: 130.0, height: 10.0),
      CGRect(x: 170.0, y: 1210.0, width: 130.0, height: 30.0),
      CGRect(x: 20.0, y: 1270.0, width: 130.0, height: 25.0),
      CGRect(x: 20.0, y: 1325.0, width: 80.0, height: 15.0),
      CGRect(x: 20.0, y: 1370.0, width: 280.0, height: 20.0),
      CGRect(x: 20.0, y: 1420.0, width: 80.0, height: 10.0),
      CGRect(x: 20.0, y: 1460.0, width: 40.0, height: 15.0),
      CGRect(x: 80.0, y: 1460.0, width: 40.0, height: 15.0),
      CGRect(x: 140.0, y: 1460.0, width: 40.0, height: 25.0),
      CGRect(x: 200.0, y: 1460.0, width: 40.0, height: 35.0),
      CGRect(x: 260.0, y: 1460.0, width: 40.0, height: 30.0),
      CGRect(x: 0.0, y: 1525.0, width: 320.0, height: 15.0),
    ]
    let expectedHeaderFrames2: [CGRect] = [
      CGRect(x: 10.0, y: 10.0, width: 300.0, height: 50.0),
    ]
    let expectedHeaderFrames3: [CGRect] = [
      CGRect(x: 10.0, y: 1130.0, width: 300.0, height: 70.0),
    ]
    let expectedFooterFrames2 = [CGRect]()
    let expectedFooterFrames3: [CGRect] = [
      CGRect(x: 10.0, y: 1060.0, width: 300.0, height: 50.0),
      CGRect(x: 10.0, y: 1550.0, width: 300.0, height: 70.0),
    ]
    let expectedBackgroundFrames2: [CGRect] = [
      CGRect(x: 10.0, y: 10.0, width: 300.0, height: 1110.0),
    ]
    let expectedBackgroundFrames3: [CGRect] = [
      CGRect(x: 10.0, y: 10.0, width: 300.0, height: 1110.0),
      CGRect(x: 10.0, y: 1130.0, width: 300.0, height: 500.0),
    ]

    checkExpectedFrames(
      expectedItemFrames0: expectedItemFrames2,
      expectedItemFrames1: expectedItemFrames3,
      expectedHeaderFrames0: expectedHeaderFrames2,
      expectedHeaderFrames1: expectedHeaderFrames3,
      expectedFooterFrames0: expectedFooterFrames2,
      expectedFooterFrames1: expectedFooterFrames3,
      expectedBackgroundFrames0: expectedBackgroundFrames2,
      expectedBackgroundFrames1: expectedBackgroundFrames3)
  }

  // MARK: Private

  private var modelState: ModelState!

  private let visibleRect0 = CGRect(x: 0, y: 0, width: 320, height: 500)
  private let visibleRect1 = CGRect(x: 0, y: 500, width: 320, height: 2000)

  private lazy var headerModel0 = HeaderModel(
    heightMode: .static(height: 50),
    height: 50)
  private lazy var sizeModesAndHeights0: [(sizeMode: MagazineLayoutItemSizeMode, height: CGFloat)] = [
    (MagazineLayoutItemSizeMode(
      widthMode: .fullWidth(respectsHorizontalInsets: true),
      heightMode: .static(height: 20)),
     20),

    (MagazineLayoutItemSizeMode(
      widthMode: .fullWidth(respectsHorizontalInsets: false),
      heightMode: .dynamic),
     15),

    (MagazineLayoutItemSizeMode(widthMode: .halfWidth, heightMode: .static(height: 10)), 10),
    (MagazineLayoutItemSizeMode(widthMode: .halfWidth, heightMode: .static(height: 30)), 30),
    (MagazineLayoutItemSizeMode(
      widthMode: .halfWidth,
      heightMode: .dynamicAndStretchToTallestItemInRow),
     25),

    (MagazineLayoutItemSizeMode(widthMode: .thirdWidth, heightMode: .dynamic), 15),
    (MagazineLayoutItemSizeMode(
      widthMode: .thirdWidth,
      heightMode: .dynamicAndStretchToTallestItemInRow),
     10),

    (MagazineLayoutItemSizeMode(widthMode: .fourthWidth, heightMode: .static(height: 35)), 15),
    (MagazineLayoutItemSizeMode(widthMode: .fourthWidth, heightMode: .dynamic), 15),
    (MagazineLayoutItemSizeMode(widthMode: .fourthWidth, heightMode: .dynamic), 25),
    (MagazineLayoutItemSizeMode(widthMode: .fourthWidth, heightMode: .dynamic), 35),

    (MagazineLayoutItemSizeMode(widthMode: .fifthWidth, heightMode: .dynamic), 30),
  ]
  private lazy var footerModel0 = FooterModel(
    heightMode: .static(height: 50),
    height: 50)

  private lazy var headerModel1 = HeaderModel(heightMode: .dynamic, height: 70)
  private lazy var sizeModesAndHeights1: [(sizeMode: MagazineLayoutItemSizeMode, height: CGFloat)] = [
    (MagazineLayoutItemSizeMode(widthMode: .halfWidth, heightMode: .static(height: 10)), 10),
    (MagazineLayoutItemSizeMode(widthMode: .halfWidth, heightMode: .static(height: 30)), 30),
    (MagazineLayoutItemSizeMode(widthMode: .halfWidth, heightMode: .static(height: 25)), 25),

    (MagazineLayoutItemSizeMode(widthMode: .thirdWidth, heightMode: .static(height: 15)), 15),

    (MagazineLayoutItemSizeMode(
      widthMode: .fullWidth(respectsHorizontalInsets: true),
      heightMode: .static(height: 20)),
     20),

    (MagazineLayoutItemSizeMode(widthMode: .thirdWidth, heightMode: .static(height: 10)), 10),

    (MagazineLayoutItemSizeMode(widthMode: .fifthWidth, heightMode: .static(height: 35)), 15),
    (MagazineLayoutItemSizeMode(widthMode: .fifthWidth, heightMode: .static(height: 15)), 15),
    (MagazineLayoutItemSizeMode(widthMode: .fifthWidth, heightMode: .static(height: 25)), 25),
    (MagazineLayoutItemSizeMode(widthMode: .fifthWidth, heightMode: .static(height: 35)), 35),
    (MagazineLayoutItemSizeMode(widthMode: .fifthWidth, heightMode: .static(height: 30)), 30),

    (MagazineLayoutItemSizeMode(
      widthMode: .fullWidth(respectsHorizontalInsets: false),
      heightMode: .static(height: 15)),
     15),
  ]
  private lazy var footerModel1 = FooterModel(heightMode: .dynamic, height: 70)

  private func checkExpectedFrames(
    expectedItemFrames0: [CGRect],
    expectedItemFrames1: [CGRect],
    expectedHeaderFrames0: [CGRect],
    expectedHeaderFrames1: [CGRect],
    expectedFooterFrames0: [CGRect],
    expectedFooterFrames1: [CGRect],
    expectedBackgroundFrames0: [CGRect],
    expectedBackgroundFrames1: [CGRect])
  {
    XCTAssert(
      FrameHelpers.expectedFrames(
        expectedItemFrames0,
        match: modelState.itemFrameInfo(forItemsIn: visibleRect0)),
      "Item frames are incorrect")
    XCTAssert(
      FrameHelpers.expectedFrames(
        expectedItemFrames1,
        match: modelState.itemFrameInfo(forItemsIn: visibleRect1)),
      "Item frames are incorrect")
    XCTAssert(
      FrameHelpers.expectedFrames(
        (expectedItemFrames0 + expectedItemFrames1).removingDuplicates(),
        matchItemFramesInSectionIndexRange: 0..<modelState.numberOfSections(.afterUpdates),
        modelState: modelState),
      "Item frames are incorrect")

    XCTAssert(
      FrameHelpers.expectedFrames(
        expectedHeaderFrames0,
        match: modelState.headerFrameInfo(forHeadersIn: visibleRect0)),
      "Header frames are incorrect")
    XCTAssert(
      FrameHelpers.expectedFrames(
        expectedHeaderFrames1,
        match: modelState.headerFrameInfo(forHeadersIn: visibleRect1)),
      "Header frames are incorrect")
    XCTAssert(
      FrameHelpers.expectedFrames(
        (expectedHeaderFrames0 + expectedHeaderFrames1).removingDuplicates(),
        matchHeaderFramesInSectionIndexRange: 0..<modelState.numberOfSections(.afterUpdates),
        modelState: modelState),
      "Header frames are incorrect")

    XCTAssert(
      FrameHelpers.expectedFrames(
        expectedFooterFrames0,
        match: modelState.footerFrameInfo(forFootersIn: visibleRect0)),
      "Footer frames are incorrect")
    XCTAssert(
      FrameHelpers.expectedFrames(
        expectedFooterFrames1,
        match: modelState.footerFrameInfo(forFootersIn: visibleRect1)),
      "Footer frames are incorrect")
    XCTAssert(
      FrameHelpers.expectedFrames(
        (expectedFooterFrames0 + expectedFooterFrames1).removingDuplicates(),
        matchFooterFramesInSectionIndexRange: 0..<modelState.numberOfSections(.afterUpdates),
        modelState: modelState),
      "Footer frames are incorrect")

    XCTAssert(
      FrameHelpers.expectedFrames(
        expectedBackgroundFrames0,
        match: modelState.backgroundFrameInfo(forBackgroundsIn: visibleRect0)),
      "Background frames are incorrect")
    XCTAssert(
      FrameHelpers.expectedFrames(
        expectedBackgroundFrames1,
        match: modelState.backgroundFrameInfo(forBackgroundsIn: visibleRect1)),
      "Background frames are incorrect")
    XCTAssert(
      FrameHelpers.expectedFrames(
        (expectedBackgroundFrames0 + expectedBackgroundFrames1).removingDuplicates(),
        matchBackgroundFramesInSectionIndexRange: 0..<modelState.numberOfSections(.afterUpdates),
        modelState: modelState),
      "Background frames are incorrect")
  }

}
